# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - AddLinkToPage

    Add a link using a form in the page, getting details of the linked document
    and inserting them with the link itself.

    @copyright: 2010, 2013 Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL, see COPYING for details.
"""

Dependencies = ['pages']

from MoinMoin.action import ActionBase
from MoinMoin.PageEditor import PageEditor
from MoinMoin.wikiutil import escape
from MoinSupport import ActionSupport, escattr
import urllib
import re

# Page parsing.

macro_pattern = re.compile(
    ur'^(?P<leading>.*?)'           # leading text on the line
    ur'<<AddLinkToPage\('           # macro prologue
    ur'(?P<identifier>[^\s,)]+).*?' # identifier
    ur'\)>>'                        # macro epilogue
    ur'(?P<trailing>.*)$',          # trailing text on the line
    re.MULTILINE | re.UNICODE)

# Link visiting and parsing.

def attr_pattern(agroup, attrname, qgroup, vgroup):
    return (
        ur'''(?P<%s>%s)'''      # attribute name
        ur'''\s*=\s*'''         # =
        ur'''(?P<%s>['"])'''    # opening quote
        ur'''(?P<%s>.*?)'''     # value
        ur'''(?P=%s)'''         # closing quote
        % (agroup, attrname, qgroup, vgroup, qgroup)
        )

meta_pattern = re.compile(
    ur'<meta'
    ur'([^>]*?'
    ur'('
    ur'(' + attr_pattern("nattr", "name", "nquote", "name") + ')'
    ur'|(' + attr_pattern("cattr", "content", "cquote", "content") + ')'
    ur')'
    ur')*'
    ur'[^>]*?>',
    re.MULTILINE | re.DOTALL)

title_pattern = re.compile(
    ur'<(?P<tag>title|h\d)(\s.*?)?>'
    ur'(?P<title>.*?)'
    ur'</(?P=tag)>',
    re.MULTILINE | re.DOTALL)

paragraph_pattern = re.compile(
    ur'<p(\s.*?)?>'
    ur'(?P<text>.*?)'
    ur'(?=<p(\s.*?)?>|</p>)',
    re.MULTILINE | re.DOTALL)

tag_pattern = re.compile(
    ur'<.*?>',
    re.MULTILINE | re.DOTALL)

def get_text(s):
    try:
        return unicode(s, "utf-8")
    except UnicodeError:
        return unicode(s, "iso-8859-1")

def get_link_info(link):

    "Get information from the given 'link'."

    # Insist on remote URLs!
    # NOTE: This could probably be done better.

    if not link.startswith("http:"):
        return None

    try:
        f = urllib.urlopen(link)
    except IOError:
        return None

    try:
        s = get_text(f.read())

        # Look for metadata.

        title = None
        intro = None

        for meta_match in meta_pattern.finditer(s):
            name = meta_match.group("name")
            content = meta_match.group("content")
            if name == "title":
                title = content
            elif name == "description":
                intro = content

        if title and intro:
            return title, intro

        # Look for titles/headings and accompanying text.

        first_title = ""

        for title_match in title_pattern.finditer(s):
            title = title_match.group("title").strip()
            start, end = title_match.span()

            if not first_title:
                first_title = title

            for intro_match in paragraph_pattern.finditer(s[end:]):
                intro = get_flattened_content(intro_match.group("text")).strip()
                if intro:
                    return title, intro
    finally:
        f.close()

    return first_title, u""

def get_flattened_content(s):

    "Get HTML or XHTML without the tags."

    l = []
    last = 0
    for match in tag_pattern.finditer(s):
        start, end = match.span()
        l.append(s[last:start])
        last = end
    l.append(s[last:])
    return "".join(l).replace("\n", " ")

# Action class and supporting functions.

class AddLinkToPage(ActionBase, ActionSupport):

    "An action adding links to pages."

    def get_form_html(self, buttons_html):
        _ = self._
        request = self.request
        page = self.page
        form = self.get_form()

        identifier = form.get("identifier", [""])[0]
        identifier_list = []

        if not identifier:

            # Show all macro identifiers by parsing the page.

            page_body = page.get_raw_body()

            for match in macro_pattern.finditer(page_body):
                found_identifier = match.group("identifier")
                identifier_list.append('<option value="%s">%s</option>' % (
                    escattr(found_identifier), escape(found_identifier)))

        link = form.get("link", [""])[0]
        insert_before = form.get('insert_before', [""])[0]
        title = ""
        introduction = ""

        # Acquire information from the link.

        if link:
            link_info = get_link_info(link)

            # NOTE: Perhaps show a message upon success/failure.

            if link_info is not None:
                title, introduction = link_info

        d = {
            "buttons_html"          : buttons_html,
            "identifiers_label"     : escape(_("Add to position...")),
            "identifier"            : escattr(identifier),
            "insert_before"         : insert_before and "true" or "",
            "link"                  : escattr(link),
            "title"                 : escattr(title),
            "intro"                 : escattr(introduction),
            "url_label"             : escape(_("URL")),
            "title_label"           : escape(_("Title")),
            "intro_label"           : escape(_("Introduction")),
            "description_label"     : escape(_("Description")),
            "preview_label"         : escape(_("Preview link")),
            }

        # Given an identifier, preserve the state in a hidden field.

        if not identifier_list:
            html = '''
  <input type="hidden" name="identifier" value="%(identifier)s" />''' % d
        else:
            html = ''

        # Start the rest of the form.

        html += u'''
  <input type="hidden" name="insert_before" value="%(insert_before)s" />
  <table>
    <tr>
      <td class="label">%(url_label)s</td>
      <td><input type="text" name="link" value="%(link)s" size="40" /></td>
    </tr>''' % d

        # Given no identifier, show the choice of insertion positions.

        if identifier_list:
            html += '''
    <tr>
      <td class="label">%(identifiers_label)s</td>
      <td>
        <select name="identifier">''' % d

            for identifier_option in identifier_list:
                html += '''
          %s''' % identifier_option

            html += '''
        </select>
      </td>
    </tr>'''

        # Finish the form.

        html += '''
    <tr>
      <td class="label">%(title_label)s</td>
      <td><input type="text" name="title" value="%(title)s" size="40" /></td>
    </tr>
    <tr>
      <td class="label">%(intro_label)s</td>
      <td><textarea name="introduction" cols="40" rows="3">%(intro)s</textarea></td>
    </tr>
    <tr>
      <td class="label">%(description_label)s</td>
      <td><input type="text" name="description" size="40" /></td>
    </tr>
    <tr>
      <td></td>
      <td class="buttons">
        <input type="submit" value="%(preview_label)s" />%(buttons_html)s
      </td>
    </tr>
  </table>''' % d

        return html

    def do_action(self):

        "Create the new event."

        _ = self._
        form = self.get_form()

        # If no title exists in the request, an error message is returned.

        identifier = form.get("identifier", [None])[0]
        link = form.get("link", [None])[0]

        if not identifier:
            return 0, _("No identifier specified.")

        if not link:
            return 0, _("No link specified.")

        return self.add_link_to_page(identifier, link)

    def render_success(self, msg, msgtype=None):

        """
        Render neither 'msg' nor 'msgtype' since redirection should occur
        instead.
        NOTE: msgtype is optional because MoinMoin 1.5.x does not support it.
        """

        pass

    def add_link_to_page(self, identifier, link):

        """
        For the macro with the given 'identifier', add 'link' to the current
        page.
        """

        _ = self._
        request = self.request
        page = self.page
        formatter = request.formatter
        form = self.get_form()

        # Get the link details.

        title = form.get('title', [link])[0]
        introduction = form.get('introduction', [""])[0]
        description = form.get('description', [""])[0]
        insert_before = form.get('insert_before', [""])[0]

        # Encode the link details.
        # NOTE: Should support different formatting options.

        link_details = "%s[[%s%s]]%s" % (
            introduction and ('"%s" ' % get_verbatim(introduction)) or "",
            link,
            title and ('|%s' % title) or "",
            description and (" - ''%s''" % description) or ""
            )

        # Open the page for editing.

        new_page = PageEditor(request, page.page_name)

        # Parse the page.

        page_body = page.get_raw_body()

        for match in macro_pattern.finditer(page_body):
            macro_identifier = match.group("identifier")
            leading_text = match.group("leading")
            trailing_text = match.group("trailing")
            start, end = match.span()

            # Where this identifier matches this macro's identifier, insert the
            # link details.

            if macro_identifier == identifier:
                if insert_before:
                    page_body = page_body[:start] + leading_text + link_details + trailing_text + "\n" + page_body[start:]
                else:
                    page_body = page_body[:end] + "\n" + leading_text + link_details + trailing_text + page_body[end:]

                # Save the new version of the page.

                new_page.saveText(page_body, 0)
                break

        # NOTE: Perhaps show a message upon failure.

        request.http_redirect(page.url(request))
        return 1, None

def get_verbatim(s):

    "Return 's' encoded as verbatim text."

    output = []

    for part in s.split('"'):
        if part:
            output.append('<<Verbatim("%s")>>' % part)
        else:
            output.append('')

    return '"'.join(output)

# Action function.

def execute(pagename, request):
    AddLinkToPage(pagename, request).render()

# vim: tabstop=4 expandtab shiftwidth=4
